/*
   Copyright 2013 Unicamp / CCUEC / DDS / CTMQ

   Licensed under the Apache License, Version 2.0 (the "License");
   you may not use this file except in compliance with the License.
   You may obtain a copy of the License at

       http://www.apache.org/licenses/LICENSE-2.0

   Unless required by applicable law or agreed to in writing, software
   distributed under the License is distributed on an "AS IS" BASIS,
   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   See the License for the specific language governing permissions and
   limitations under the License. 
 */
package br.unicamp.ccuec.ctmq.virusScan.impl;

import java.io.File;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.IOException;
import java.io.InputStream;
import java.net.URL;
import java.util.Properties;
import java.util.logging.Level;
import java.util.logging.Logger;

import br.unicamp.ccuec.ctmq.virusScan.spec.InfraException;
import br.unicamp.ccuec.ctmq.virusScan.spec.VirusException;
import br.unicamp.ccuec.ctmq.virusScan.spec.VirusScan;

import com.philvarner.clamavj.ScanResult;

/**
 * 
 * @author clayton@ccuec.unicamp.br (Denis Clayton Alves Ramos)
 * 
 */
public abstract class AbstractVirusScan {
	private static final String _MSG_DESENV = "\n\n*** EXECUTANDO EM MODO DESENVOLVIMENTO: motor anti-virus do servidor n\u00e3o ser\u00e1 acionado!\n\n";
	private static final String _MSG_CONFIG = " N\u00c3O ENCONTRADO, USANDO CONFIG. DEFAULT! RECOMENDA-SE CONFIG. PARA ENVIO DE E-MAIL DE ALERTA!\n";
	private static final String _MSG_EOF_CONFIG = ". Usando configura\u00e7\u00f5es default, por\u00e9m sem endere\u00e7os de e-mail para alerta!\n";

	private static final Logger _LOGGER = Logger.getLogger(AbstractVirusScan.class
			.getName());

	private static final String _PROPERTIES_PATH = "virusScan-conf.properties";

	private static Properties props = null;	

	/**
	 * relacao de e-mails para aviso de alerta de anti-virus nao respondendo
	 */
	protected static String[] alertMailsTo = null;
	protected static String[] alertMailsCc = null;

	/**
	 * config default
	 */
	protected static String host = "127.0.0.1";
	protected static int port = 3310;
	protected static int timeout = 120000;
	protected static int provider = 1;

	protected static final String _MSG_ERROR_INFRA = "\n\nPROBLEMA INFRA-ESTRUTURA ANTI-V\u00cdRUS: ";

	static {
		if (Boolean.getBoolean("skipVirusScan")) {
			provider = -1;

			_LOGGER.log(Level.INFO, _MSG_DESENV);
		} else {
			loadConfig();
			autoTest();
		}
	}

	protected static void loadConfig() {
		InputStream is = null;
		try {
			URL path = Thread.currentThread().getContextClassLoader()
					.getResource(_PROPERTIES_PATH);
			if (path == null) {
				throw new FileNotFoundException();
			}
			is = path.openStream();

			props = new Properties();
			props.load(is); 

			String strHost = props.getProperty("host");
			if (strHost != null && !"127.0.0.1".equals(strHost.trim())) {
				host = strHost.trim();
			}

			String strPort = props.getProperty("port");
			if (strPort != null && !"3310".equals(strPort.trim())) {
				port = new Integer(strPort.trim()).intValue();
			}

			String strTimeout = props.getProperty("timeout");
			if (strTimeout != null && !"120000".equals(strTimeout.trim())) {
				timeout = new Integer(strTimeout.trim()).intValue();
			}

			String strTo = props.getProperty("para");
			if (strTo != null) {
				alertMailsTo = strTo.trim().split(",");
			}

			String strCc = props.getProperty("cc");
			if (strCc != null) {
				alertMailsCc = strCc.trim().split(",");
			}

			String strProvider = props.getProperty("anti-virus-provider");
			if (strProvider != null && !"1".equals(strProvider.trim())) {
				provider = new Integer(strProvider.trim()).intValue();
			}

			_LOGGER.log(Level.INFO, "\n\nANTI-VIRUS CONFIG.\nhost = " + host
					+ ", port = " + port + ", timeout = " + timeout + ", to = "
					+ alertMailsTo.length + ", provider = " + provider);
		}
		catch (FileNotFoundException e) {
			_LOGGER.log(Level.SEVERE, "\n\n" + _PROPERTIES_PATH + _MSG_CONFIG);
			e.printStackTrace();
		}
		catch (IOException e) {
			_LOGGER.log(Level.SEVERE, "\n\n ERRO LEITURA " + _PROPERTIES_PATH
					+ _MSG_EOF_CONFIG);
			e.printStackTrace();
		}
		finally {
			try {
				if (is != null) {
					is.close();
				}
			} catch (IOException e) {
				e.printStackTrace();
			}
		}
	}

	/**
	 * teste conexao / ping motor anti-virus
	 */
	protected static void autoTest() {
		RunAutoTest autoTestRun = new RunAutoTest(alertMailsTo, alertMailsCc,
				_MSG_ERROR_INFRA, _LOGGER);

		Thread thread = new Thread(autoTestRun);
		thread.start();
	}

	protected static void printScanStatus(ScanResult result, String... args) {
		StringBuilder stb = new StringBuilder();
		stb.append("\n====[RESULTADO]====\n");
		stb.append("\n " + args[0] + " "
				+ ((args.length > 1) ? (String) args[1] : ""));
		stb.append("\nstring...: " + result.getResult());
		stb.append("\nsignature: " + result.getSignature());
		stb.append("\nstatus...: " + result.getStatus());
		stb.append("\nexception: " + result.getException());
		stb.append("\n===================\n\n");
		_LOGGER.log(Level.WARNING, stb.toString());
	}

	/**
	 * 
	 * @return @see br.unicamp.ccuec.ctmq.virusScan.spec.VirusScan
	 */
	public static VirusScan getScanner() {
		VirusScan scanner = null;

		switch (provider) {
		case -1:
			scanner = new VirusScanFakeImpl();
			break;

		case 1:
			scanner = new VirusScanClamavis();
			break;			

		default:
			scanner = new VirusScanClamavis();
			break;
		}
		return scanner;
	}

	// =====================[ metodos para teste manual ]=========================
	private static void testVirusFromFile(VirusScan scanner, String path) {
		if (scanner == null) {
			_LOGGER.log(Level.SEVERE, "\nnumero-antivirus-provider n\u00e3o corresponde a nenhuma implementa\u00e7\u00e3o!");
		}
		_LOGGER.log(Level.INFO, "\n\n...scanning...");

		File file = null;
		FileInputStream fis = null;

		try {
			file = new File(path);
			fis = new FileInputStream(file);

			if (scanner.scan(fis)) {
				_LOGGER.log(Level.WARNING, "\n arquivo " + path
						+ " CHECADO E LIMPO!!!\n\n");
			}
		}
		catch (FileNotFoundException e) {
			_LOGGER.log(Level.WARNING, "\n arquivo " + path
					+ " N\u00c3O PODE SER ESCANEADO!");
			e.printStackTrace();
		}
		catch (VirusException e) {
			_LOGGER.log(Level.WARNING, "\n arquivo " + path
					+ " VIRUS DETECTADO!!!\n\n");
		}
		catch (InfraException e) {
			_LOGGER.log(Level.WARNING, _MSG_ERROR_INFRA
					+ " processamento arquivo " + path);
			e.printStackTrace();
		}
		finally {
			try {
				if (fis != null) {
					fis.close();
				}
			} catch (IOException ioe) {
				_LOGGER.log(Level.WARNING, path
						+ " N\u00c3O DESCARREGADO DA MEM\u00d3RIA!");
				ioe.printStackTrace();
			}
		}
	}

	public static void main(String[] args) {
		if (args != null && args.length > 0) {
			if (args.length > 1) {
				provider = new Integer(args[1].trim()).intValue();
			}

			VirusScan scanner = getScanner();

			testVirusFromFile(scanner, args[0]);
		} else {
			_LOGGER.log(Level.INFO,
					"\nFORMAS DE USO: \n\t(a) java -cp mail-1.4.5.jar -jar virus-scan-1.0.0-SNAPSHOT.jar path-arquivo\n"
					+ "\n\t(b) java -cp mail-1.4.5.jar -jar virus-scan-1.0.0-SNAPSHOT.jar path-arquivo numero-anti-virus-provider\n");
		}
	}
}